<?php

namespace Drupal\bpmn_io\Plugin\modeler_api_modeler;

use Drupal\bpmn_io\Service\Parser;
use Drupal\bpmn_io\Service\PrepareComponents;
use Drupal\Component\Utility\Random;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\OpenOffCanvasDialogCommand;
use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Form\FormBuilder;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\bpmn_io\Form\Modeler as ModelerForm;
use Drupal\bpmn_io\Service\TokenBrowserService;
use Drupal\modeler_api\Api;
use Drupal\modeler_api\Attribute\Modeler;
use Drupal\modeler_api\Form\Wrapper;
use Drupal\modeler_api\Plugin\modeler_api_model_owner\ModelOwnerInterface;
use Drupal\modeler_api\Plugin\modeler_api_modeler\ModelerBase;
use Drupal\modeler_api\Plugin\modeler_api_modeler\ModelerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the Modeler API.
 */
#[Modeler(
  id: "bpmn_io",
  label: new TranslatableMarkup("BPMN.iO"),
  description: new TranslatableMarkup("BPMN modeler with a feature-rich UI.")
)]
class BpmnIo extends ModelerBase {

  public const array SUPPORTED_COMPONENT_TYPES = [
    Api::COMPONENT_TYPE_START => 'bpmn:StartEvent',
    Api::COMPONENT_TYPE_LINK => 'bpmn:SequenceFlow',
    Api::COMPONENT_TYPE_ELEMENT => 'bpmn:Task',
    Api::COMPONENT_TYPE_GATEWAY => 'bpmn:Gateway',
    Api::COMPONENT_TYPE_SUBPROCESS => 'bpmn:SubProcess',
  ];

  public const array MAP_COMPONENT_TYPES = [
    'bpmn:StartEvent' => 'StartEvent',
    'bpmn:SequenceFlow' => 'SequenceFlow',
    'bpmn:Task' => 'Task',
    'bpmn:Gateway' => 'ExclusiveGateway',
    'bpmn:SubProcess' => 'SubProcess',
  ];

  /**
   * Dependency Injection container.
   *
   * Used for getter injection.
   *
   * @var \Symfony\Component\DependencyInjection\ContainerInterface|null
   */
  protected ?ContainerInterface $container;

  /**
   * The form builder.
   *
   * @var \Drupal\Core\Form\FormBuilder
   */
  protected FormBuilder $formBuilder;

  /**
   * The token browser service.
   *
   * @var \Drupal\bpmn_io\Service\TokenBrowserService
   */
  protected TokenBrowserService $tokenBrowserService;

  /**
   * The modeler API.
   *
   * @var \Drupal\modeler_api\Api
   */
  protected Api $modelerApi;

  /**
   * The prepare components service.
   *
   * @var \Drupal\bpmn_io\Service\PrepareComponents
   */
  protected PrepareComponents $prepareComponents;

  /**
   * The BPMN parser.
   *
   * @var \Drupal\bpmn_io\Service\Parser
   */
  protected Parser $parser;

  /**
   * Get Dependency Injection container.
   *
   * @return \Symfony\Component\DependencyInjection\ContainerInterface
   *   Current Dependency Injection container.
   */
  protected function getContainer(): ContainerInterface {
    if (!isset($this->container)) {
      // @phpstan-ignore-next-line
      $this->container = \Drupal::getContainer();
    }
    return $this->container;
  }

  /**
   * Get the form builder.
   *
   * @return \Drupal\Core\Form\FormBuilder
   *   The form builder.
   */
  protected function formBuilder(): FormBuilder {
    if (!isset($this->formBuilder)) {
      $this->formBuilder = $this->getContainer()->get('form_builder');
    }
    return $this->formBuilder;
  }

  /**
   * Get the token browser service.
   *
   * @return \Drupal\bpmn_io\Service\TokenBrowserService
   *   The token browser service.
   */
  protected function tokenBrowserService(): TokenBrowserService {
    if (!isset($this->tokenBrowserService)) {
      $this->tokenBrowserService = $this->getContainer()->get('bpmn_io.token_browser');
    }
    return $this->tokenBrowserService;
  }

  /**
   * Get the modeler API.
   *
   * @return \Drupal\modeler_api\Api
   *   The modeler API.
   */
  protected function modelerApi(): Api {
    if (!isset($this->modelerApi)) {
      $this->modelerApi = $this->getContainer()->get('modeler_api.service');
    }
    return $this->modelerApi;
  }

  /**
   * Get the prepare components service.
   *
   * @return \Drupal\bpmn_io\Service\PrepareComponents
   *   The prepare components service.
   */
  protected function prepareComponents(): PrepareComponents {
    if (!isset($this->prepareComponents)) {
      $this->prepareComponents = $this->getContainer()->get('bpmn_io.prepare_components');
    }
    return $this->prepareComponents;
  }

  /**
   * Get the BPMN parser.
   *
   * @return \Drupal\bpmn_io\Service\Parser
   *   The BPMN parser.
   */
  protected function parser(): Parser {
    if (!isset($this->parser)) {
      $this->parser = $this->getContainer()->get('bpmn_io.parser');
    }
    return $this->parser;
  }

  /**
   * {@inheritdoc}
   */
  public function getRawFileExtension(): ?string {
    return 'xml';
  }

  /**
   * {@inheritdoc}
   */
  public function isEditable(): bool {
    return TRUE;
  }

  /**
   * Returns a render array with everything required for model editing.
   *
   * @return array
   *   The render array.
   */
  public function edit(ModelOwnerInterface $owner, string $id, string $data, bool $isNew = FALSE): array {
    $form = $this->formBuilder()->getForm(ModelerForm::class, $owner, $id);
    if (isset($form['gin_sidebar'])) {
      $form['gin_sidebar']['property_panel'] = ['#markup' => '<div class="property-panel"></div>'];
      $form['gin_sidebar']['token_browser'] = $this->tokenBrowserService()->getTokenBrowserMarkup();
      $extras = '';
    }
    else {
      $extras = '<div class="property-panel in-canvas"></div>';
      $form['actions']['tb'] = $this->tokenBrowserService()->getTokenBrowserMarkup();
    }
    $mutuallySupported = array_intersect_key(self::SUPPORTED_COMPONENT_TYPES, $owner->supportedOwnerComponentTypes());
    $classes = [];
    foreach ($mutuallySupported as $item => $value) {
      $classes[] = match ($item) {
        Api::COMPONENT_TYPE_START => 'support-events',
        Api::COMPONENT_TYPE_SUBPROCESS => 'support-subprocesses',
        Api::COMPONENT_TYPE_ELEMENT => 'support-tasks',
        Api::COMPONENT_TYPE_LINK => 'support-links',
        // @phpstan-ignore-next-line
        Api::COMPONENT_TYPE_GATEWAY => 'support-gateways',
        default => '',
      };
    }
    return [
      '#type' => 'container',
      '#attributes' => [
        'id' => 'bpmn-io',
        'class' => $classes,
      ],
      'canvas' => [
        '#prefix' => '<div class="canvas" role="application" aria-label="BPMN Canvas"></div>' . $extras,
      ],
      'form' => $form,
      '#attached' => [
        'library' => [
          'bpmn_io/ui',
        ],
        'drupalSettings' => [
          'bpmn_io' => [
            'id' => $id,
            'is_new' => $isNew,
            'modeler' => 'bpmn_io',
            'bpmn' => $data,
            'templates' => $this->prepareComponents()->getTemplates($owner),
          ],
        ],
      ],
    ];
  }

  /**
   * Returns a render array with everything required for model editing.
   *
   * @return array
   *   The render array.
   */
  public function convert(ModelOwnerInterface $owner, ConfigEntityInterface $model): array {
    $reload = FALSE;
    if (!in_array($owner->getModelerId($model), ['bpmn_io', 'fallback'])) {
      // Let the modeler clone the model entity.
      $model = $owner->clone($model);
      $reload = $owner->configEntityBasePath() !== NULL;
    }
    $owner->setModelerId($model, 'bpmn_io');
    $type = $model->getEntityTypeId();

    $id = '';
    $emptyModelData = $this->prepareEmptyModelData($id);
    if ($model->isNew()) {
      $model->set('id', $id);
    }
    else {
      $emptyModelData = str_replace($id, $model->id(), $emptyModelData);
    }

    // Provide enough mappings so that the Javascript-class can do its thing.
    $build = $this->edit($owner, $model->id(), $emptyModelData, $model->isNew());
    $build['canvas']['#suffix'] = '<div class="convert-overlay">' . $this->t('Conversion in progress ...') . '</div>';
    $build['#attached']['library'][] = 'bpmn_io/convert';
    $build['#attached']['drupalSettings']['bpmn_io_convert'] = [
      'reload' => $reload,
      'metadata' => [
        'name' => $model->label(),
        'version' => $model->get('version'),
      ],
    ];
    if ($owner->configEntityBasePath() !== NULL) {
      $build['#attached']['drupalSettings']['bpmn_io_convert']['metadata']['redirect_url'] = Url::fromRoute(
        'entity.' . $type . '.edit_form',
        [$type => $model->id()],
        ['absolute' => TRUE]
      )->toString();
    }

    $components = $owner->usedComponents($model);
    $supportedOwnerComponentTypes = $owner->supportedOwnerComponentTypes();
    $elements = $bpmnMapping = $templateMapping = [];
    foreach ($components as $component) {
      if (!isset(self::SUPPORTED_COMPONENT_TYPES[$component->getType()])) {
        // @todo Log this missing component type.
        continue;
      }
      $successors = [];
      foreach ($component->getSuccessors() as $successor) {
        $successors[] = [
          'id' => $successor->getId(),
          'condition' => $successor->getConditionId(),
        ];
      }
      $elements[$component->getId()] = [
        'plugin' => $component->getPluginId(),
        'label' => $component->getLabel(),
        'configuration' => $component->getConfiguration(),
        'successors' => $successors,
        'parentId' => $component->getParentId(),
      ];
      $bpmnMapping[$component->getId()] = self::MAP_COMPONENT_TYPES[self::SUPPORTED_COMPONENT_TYPES[$component->getType()]];
      $templateMapping[$component->getId()] = $supportedOwnerComponentTypes[$component->getType()];
    }
    $build['#attached']['drupalSettings']['bpmn_io_convert']['elements'] = $elements;
    $build['#attached']['drupalSettings']['bpmn_io_convert']['bpmn_mapping'] = $bpmnMapping;
    $build['#attached']['drupalSettings']['bpmn_io_convert']['template_mapping'] = $templateMapping;

    return $build;
  }

  /**
   * {@inheritdoc}
   */
  public function generateId(): string {
    $random = new Random();
    return 'Process_' . $random->name(7);
  }

  /**
   * {@inheritdoc}
   */
  public function enable(ModelOwnerInterface $owner): ModelerInterface {
    $this->parser()->enable($owner);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function disable(ModelOwnerInterface $owner): ModelerInterface {
    $this->parser()->disable($owner);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function clone(ModelOwnerInterface $owner, string $id, string $label): ModelerInterface {
    $this->parser()->clone($owner, $id, $label);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function prepareEmptyModelData(string &$id): string {
    $id = $this->generateId();
    $emptyBpmn = file_get_contents($this->extensionPathResolver->getPath('module', 'bpmn_io') . DIRECTORY_SEPARATOR . 'data' . DIRECTORY_SEPARATOR . 'empty.bpmn');
    return str_replace([
      'SIDPLACEHOLDER1',
      'SIDPLACEHOLDER2',
      'IDPLACEHOLDER',
    ], [
      'sid-' . $this->uuid->generate(),
      'sid-' . $this->uuid->generate(),
      $id,
    ], $emptyBpmn);
  }

  /**
   * {@inheritdoc}
   */
  public function getId(): string {
    return $this->parser()->getId();
  }

  /**
   * {@inheritdoc}
   */
  public function getLabel(): string {
    return $this->parser()->getLabel();
  }

  /**
   * {@inheritdoc}
   */
  public function getTags(): array {
    return $this->parser()->getTags();
  }

  /**
   * {@inheritdoc}
   */
  public function getChangelog(): array {
    return $this->parser()->getChangelog();
  }

  /**
   * {@inheritdoc}
   */
  public function getDocumentation(): string {
    return $this->parser()->getDocumentation();
  }

  /**
   * {@inheritdoc}
   */
  public function getStatus(): bool {
    return $this->parser()->getStatus();
  }

  /**
   * {@inheritdoc}
   */
  public function getVersion(): string {
    return $this->parser()->getVersion();
  }

  /**
   * {@inheritdoc}
   */
  public function parseData(ModelOwnerInterface $owner, string $data): void {
    parent::parseData($owner, $data);
    $this->parser()->setData($owner, $data);
  }

  /**
   * {@inheritdoc}
   */
  public function readComponents(): array {
    return $this->parser()->getComponents();
  }

  /**
   * {@inheritdoc}
   */
  public function updateComponents(ModelOwnerInterface $owner): bool {
    $changed = $this->parser()->updateComponents($owner, $this->prepareComponents()->getTemplates($owner));
    $this->parseData($owner, $this->parser()->getData());
    return $changed;
  }

  /**
   * {@inheritdoc}
   */
  public function configForm(ModelOwnerInterface $owner): AjaxResponse {
    $response = new AjaxResponse();
    parse_str($this->request->getContent(), $data);
    $key = array_search($data['type'], self::SUPPORTED_COMPONENT_TYPES);
    if ($plugin = $owner->ownerComponent($key, $data['pluginId'], $data['config'] ?? [])) {
      $form = $owner->buildConfigurationForm($plugin, $data['entityId'], $data['isNew'] === 'true');
      if (!isset($form['#title'])) {
        if (($data['config']['label'] ?? '') !== '') {
          $form['#title'] = $data['config']['label'];
        }
        else {
          $form['#title'] = $plugin->getPluginDefinition()['label'] ?? $plugin->getPluginDefinition()['name'] ?? $plugin->getPluginId();
        }
      }
    }
    else {
      $form = [
        '#title' => $this->t('Unknown component'),
      ];
    }
    $form = $this->formBuilder()->getForm(Wrapper::class, $form);
    $form['#attached']['library'][] = 'core/drupal.dialog.off_canvas';

    $options['width'] = $data['width'] === '' ? OpenOffCanvasDialogCommand::DEFAULT_DIALOG_WIDTH : $data['width'];
    $response->addCommand(new OpenOffCanvasDialogCommand($form['#title'], $form, $options));
    return $response;
  }

}
